from numpy import *
from numpy.linalg import LinAlgError
from numpy.random import randint
from six import iteritems
from six.moves import range
import hashlib
from random import shuffle

from .logger import log_message

from .mask_starts import mask_starts
from .linear_algebra import BlockPlusDiagonalMatrix
from .default_parameters import default_parameters

from .numerics import (accumulate_cluster_mask_sum, compute_cluster_mean,
                       compute_covariance_matrix,
                       compute_log_p_and_assign, compute_penalties)

import time

__all__ = ['KK']

class PartitionError(Exception):
    pass


class section(object):
    def __init__(self, kk, name, *args, **kwds):
        self.kk = kk
        self.name = name
        self.args = args
        self.kwds = kwds
        if not hasattr(kk, '_section_timings_t_total'):
            kk._section_timings_t_total = {}
            kk._section_timings_num_calls = {}
        if name not in kk._section_timings_t_total:
            kk._section_timings_t_total[name] = 0.0
            kk._section_timings_num_calls[name] = 0
    def __enter__(self):
        self.kk.run_callbacks('start_'+self.name, *self.args, **self.kwds)
        self.t_start = time.time()
    def __exit__(self, type, value, traceback):
        this_time = time.time()-self.t_start
        self.kk.run_callbacks('end_'+self.name, *self.args, **self.kwds)
        if self.kk.name:
            self.kk.log('debug', 'This call: %.2f ms.' % (this_time*1000),
                        suffix='timing.'+self.name)
            return
        name = self.name
        st_t = self.kk._section_timings_t_total
        st_n = self.kk._section_timings_num_calls
        st_t[name] += this_time
        st_n[name] += 1
        mean_time = st_t[name]/st_n[name]
        self.kk.log('debug', 'This call: %.2f ms. Average: %.2f ms. Total: %.2f s. '
                             'Num calls: %d' % (this_time*1000, mean_time*1000,
                                                st_t[name], st_n[name]),
                    suffix='timing.'+self.name)


def add_slots(meth):
    def new_meth(self, *args, **kwds):
        with section(self, meth.__name__, *args, **kwds):
            res = meth(self, *args, **kwds)
        return res
    new_meth.__name__ = meth.__name__
    new_meth.__doc__ = meth.__doc__
    return new_meth


class KK(object):
    '''
    Main object used for clustering data

    This documentation is temporary, just to highlight the parts of the interface that should
    not change and can be used externally:

    * kk.special_clusters: a dict mapping special cluster names to their associated cluster index.
    * kk.num_special_clusters: the number of special clusters. This is also the cluster index of the
      first normal cluster.
    * Initialisation KK(data, use_noise_cluster=True, use_mua_cluster=True, **params)
    * Method kk.cluster_mask_starts(num_starting_clusters) will cluster from mask starts.
    * Method kk.cluster_from(clusters) will cluster from the given array of cluster assignments.
    * kk.clusters (after clustering) is the array of cluster assignments.
    * Method kk.register_callback(callback, slot=None) register a callback function that will be
      called at the given slot (see code for slot names), by default at the end of each iteration
      of the algorithm. Callback functions are normally called as f(kk), but some slots will
      defined additional arguments and keyword arguments.
    '''
    def __init__(self, data, callbacks=None, name='',
                 is_subset=False, is_copy=False,
                 map_log_to_debug=False,
                 **params):
        self.name = name
        if callbacks is None:
            callbacks = {}
        self.callbacks = callbacks
        self.data = data
        self.cluster_hashes = set()
        self.is_subset = is_subset
        self.is_copy = is_copy
        self.map_log_to_debug = map_log_to_debug
        # user parameters
        show_params = name=='' and not is_subset and not is_copy
        self.params = params
        actual_params = default_parameters.copy()
        for k, v in iteritems(params):
            if k not in default_parameters:
                raise ValueError("There is no parameter "+k)
            actual_params[k] = v
        for k, v in iteritems(actual_params):
            setattr(self, k, v)
            if show_params:
                self.log('info', '%s = %s' % (k, v), suffix='initial_parameters')
        use_noise_cluster = self.use_noise_cluster
        use_mua_cluster = self.use_mua_cluster
        self.all_params = actual_params
        # Assignment of special clusters
        self.num_special_clusters = 0
        self.first_gaussian_cluster = 0
        self.special_clusters = {}
        if use_noise_cluster:
            self.noise_cluster = self.special_clusters['noise'] = self.num_special_clusters
            self.num_special_clusters += 1
            self.first_gaussian_cluster += 1
        else:
            self.noise_cluster = -2
        if use_mua_cluster:
            self.mua_cluster = self.special_clusters['mua'] = self.num_special_clusters
            self.num_special_clusters += 1
        else:
            self.mua_cluster = -2

    def register_callback(self, callback, slot='end_iteration'):
        if slot not in self.callbacks:
            self.callbacks[slot] = []
        self.callbacks[slot].append(callback)

    def run_callbacks(self, slot, *args, **kwds):
        if slot in self.callbacks:
            for callback in self.callbacks[slot]:
                callback(self, *args, **kwds)

    def log(self, level, msg, suffix=None):
        if self.map_log_to_debug:
            level = 'debug'
        if suffix is not None:
            if self.name=='':
                name = suffix
            else:
                name = self.name+'.'+suffix
        else:
            name = self.name
        log_message(level, msg, name=name)

    def copy(self, name='kk_copy',
             use_noise_cluster=None, use_mua_cluster=None,
             **additional_params):
        if self.name:
            sep = '.'
        else:
            sep = ''
        if use_noise_cluster is None:
            use_noise_cluster = self.use_noise_cluster
        if use_mua_cluster is None:
            use_mua_cluster = self.use_mua_cluster
        params = self.params.copy()
        params.update(**additional_params)
        return KK(self.data, name=self.name+sep+name,
                  callbacks=self.callbacks,
                  use_noise_cluster=use_noise_cluster, use_mua_cluster=use_mua_cluster,
                  is_copy=True,
                  **params)

    def subset(self, spikes, name='kk_subset', **additional_params):
        newdata = self.data.subset(spikes)
        if self.name:
            sep = '.'
        else:
            sep = ''
        params = self.params.copy()
        params.update(**additional_params)
        return KK(newdata, name=self.name+sep+name,
                  callbacks=self.callbacks,
                  is_subset=True,
                  **params)

    def subset_features(self, feature_indices, name='', **additional_params):
        newdata, spikes = self.data.subset_features(feature_indices)
        if self.name:
            sep = '.'
        else:
            sep = ''
        params = self.params.copy()
        params.update(**additional_params)
        return KK(newdata, name=self.name+sep+name,
                  callbacks=self.callbacks,
                  is_subset=True,
                  **params), spikes

    def initialise_clusters(self, clusters):
        self.clusters = clusters
        self.old_clusters = -1*ones(len(self.clusters), dtype=int)
        self.reindex_clusters()

    def cluster_with_subset_schedule(self, num_starting_clusters, schedule):
        schedule = array(schedule)
        if schedule[-1]!=1:
            raise ValueError("Last value in subset schedule must be 1")
        if (diff(schedule)<=0).any():
            raise ValueError("Subset schedule must be strictly increasing")
        if (schedule<=0).any():
            raise ValueError("All subset schedule fractions must be between 0 and 1")
        spikes = arange(self.num_spikes)
        shuffle(spikes)
        I_end = unique(array(self.num_spikes*schedule, dtype=int))
        clusters = mask_starts(self.data, num_starting_clusters, self.num_special_clusters)
        clusters = clusters[spikes]
        clusters[I_end[0]:] = 0 # assign to noise, they will quickly get reassigned
        for i, i_end in enumerate(I_end):
            self.log('info', 'Subsetting stage %d of %d, %d points of %d' % (i+1, len(I_end), i_end,
                                                                             self.num_spikes))
            if i==0:
                split_first = self.split_first
            else:
                split_first = 0
            if i_end<len(spikes):
                kk_sub = self.subset(spikes[:i_end], name='', split_first=split_first,
                                     break_fraction=self.subset_break_fraction,
                                     use_noise_cluster=self.use_noise_cluster,
                                     use_mua_cluster=self.use_mua_cluster,
                                     )
            else:
                kk_sub = self.copy(name='', split_first=split_first)
                clusters_new = zeros(self.num_spikes, dtype=int)
                clusters_new[spikes] = clusters
                clusters = clusters_new
            kk_sub.cluster_from(clusters[:i_end].copy())
            clusters[:i_end] = kk_sub.clusters
        self.clusters = kk_sub.clusters
        self.reindex_clusters()

    def cluster_mask_starts(self):
        clusters = mask_starts(self.data, self.num_starting_clusters, self.num_special_clusters)
        self.cluster_from(clusters)

    def cluster_from(self, clusters, recurse=True, score_target=-inf):
        self.log('info', 'Clustering data set of %d points, %d features' % (self.data.num_spikes,
                                                                            self.data.num_features))
        self.initialise_clusters(clusters)
        return self.iterate(recurse=recurse, score_target=score_target)
    
    def prepare_for_iterate(self):
        self.full_step = True
        self.current_iteration = 0
        self.force_next_step_full = False
        self.score_history = []
    
    def iterate(self, recurse=True, score_target=-inf):        
        self.prepare_for_iterate()

        score = score_raw = score_penalty = None

        iterations_until_next_split = self.split_first
        tried_splitting_to_escape_cycle_hashes = set()

        self.log('info', 'Starting iteration 0 with %d clusters' % self.num_clusters_alive)

        while self.current_iteration<self.max_iterations:
            if self.force_next_step_full:
                self.force_next_step_full = False
                self.full_step = True
                self.log('debug', 'Next step forced to be full')
            self.log('debug', 'Starting iteration %d' % self.current_iteration)
            if self.full_step:
                self.log('debug', 'This step is a full step')
            else:
                self.log('debug', 'This step is a quick step')
            self.MEC_steps()
            self.compute_cluster_penalties()
            # only delete after a full step to simplify quick steps
            if recurse and self.full_step and self.consider_cluster_deletion:
                self.consider_deletion()
            old_score = score
            old_score_raw = score_raw
            old_score_penalty = score_penalty
            score, score_raw, score_penalty = self.compute_score()
            self.score_history.append((score, score_raw, score_penalty))
            
            clusters_changed, = (self.clusters!=self.old_clusters).nonzero()
            clusters_changed = array(clusters_changed, dtype=int)
            num_changed = len(clusters_changed)
            if num_changed and not self.full_step and self.full_step_every>1:
                # add these changed clusters to all the candidate sets
                num_candidates = 0
                max_quick_step_candidates = min(self.max_quick_step_candidates,
                    self.max_quick_step_candidates_fraction*self.num_spikes*self.num_clusters_alive)
                with section(self, 'quick_step_union'):
                    for cluster, candidates in list(self.quick_step_candidates.items()):
                        candidates = union1d(candidates, clusters_changed)
                        self.quick_step_candidates[cluster] = candidates
                        num_candidates += len(candidates)
                        if num_candidates>max_quick_step_candidates:
                            self.quick_step_candidates = dict()
                            self.force_next_step_full = True
                            if num_candidates>self.max_quick_step_candidates:
                                self.log('info', 'Ran out of storage space for quick step, try increasing '
                                                 'max_quick_step_candidates if this happens often.')
                            else:
                                self.log('debug', 'Exceeded quick step point fraction, next step '
                                                  'will be full')
                            break

            self.run_callbacks('scores', score=score, score_raw=score_raw,
                               score_penalty=score_penalty, old_score=old_score,
                               old_score_raw=old_score_raw, old_score_penalty=old_score_penalty,
                               num_changed=num_changed,
                               )

            self.current_iteration += 1

            QF_id = {True:'F', False:'Q'}[self.full_step]
            msg = 'Iteration %d%s: %d clusters, %d changed, score=%f' % (self.current_iteration, QF_id,
                                                                         self.num_clusters_alive,
                                                                         num_changed, score)

            last_step_full = self.full_step
            self.full_step = (num_changed>self.num_changed_threshold*self.num_spikes or
                              num_changed==0 or
                              self.current_iteration % self.full_step_every == 0 or
                              (old_score is not None and score > old_score))
            if not hasattr(self, 'old_log_p_best'):
                self.full_step = True

            self.reindex_clusters()

            if old_score is not None:
                msg += ' (decreased by %f)' % (old_score-score)
            self.log('info', msg)
            if old_score is not None:
                msg = 'Change in scores: raw=%f, penalty=%f, total=%f'  % (old_score_raw-score_raw,
                                                                           old_score_penalty-score_penalty,
                                                                           old_score-score)
                self.log('debug', msg)

            # Splitting logic
            iterations_until_next_split -= 1
            if num_changed==0 and last_step_full:
                self.log('info', 'No points changed and last step was full, so trying to split.')
                iterations_until_next_split = 0

            # Cycle detection/breaking
            cluster_hash = hashlib.sha1(self.clusters.view(uint8)).hexdigest()
            if cluster_hash in self.cluster_hashes and num_changed>0:
                if recurse:
                    if cluster_hash in tried_splitting_to_escape_cycle_hashes:
                        self.log('error', 'Cycle detected! Already tried attempting to break out '
                                          'by splitting, so abandoning.')
                        break
                    else:
                        self.log('warning', 'Cycle detected! Attempting to break out by splitting.')
                        iterations_until_next_split = 0
                    tried_splitting_to_escape_cycle_hashes.add(cluster_hash)
                else:
                    self.log('error', 'Cycle detected! Splitting is not enabled, so abandoning.')
                    break
            self.cluster_hashes.add(cluster_hash)

            # Try splitting
            did_split = False
            if recurse and iterations_until_next_split<=0:
                did_split = self.try_splits()
                iterations_until_next_split = self.split_every

            self.run_callbacks('end_iteration')

            if num_changed==0 and last_step_full and not did_split:
                self.log('info', 'No points changed, previous step was full and did not split, '
                                 'so finishing.')
                break

            if num_changed<self.break_fraction*self.num_spikes and last_step_full:
                self.log('info', 'Number of points changed below break fraction, so finishing.')
                break

            if score<score_target:
                self.log('info', 'Reached score target, so finishing.')
        else:
            # ran out of iterations
            self.log('info', 'Number of iterations exceeded maximum %d' % self.max_iterations)

        return score
    
    @add_slots
    def MEC_steps(self, only_evaluate_current_clusters=False):
        # eliminate any clusters with 0 members, compute the list of spikes
        # in each cluster, compute the cluster masks and allocate space for
        # covariance matrices
        self.reindex_clusters()
        # Computes the masked and unmasked indices for each cluster based on the
        # masks for each point in that cluster. Allocates space for covariance
        # matrices.
        num_clusters = self.num_clusters_alive
        num_features = self.num_features
        num_cluster_members = self.num_cluster_members
        cluster_start = self.num_special_clusters
        num_spikes = self.num_spikes

        # Weight computations
        denom = self.num_spikes+self.prior_point*(num_clusters-self.num_special_clusters)
        if self.use_noise_cluster:
            denom += self.noise_point
        if self.use_mua_cluster:
            denom += self.mua_point
        denom = float(denom)
        if self.use_noise_cluster:
            noise_weight = (num_cluster_members[self.noise_cluster]+self.noise_point)/denom
        
        # Arrays that will be used in E-step part
        if only_evaluate_current_clusters:
            self.clusters_second_best = zeros(0, dtype=int)
            self.log_p_best = empty(num_spikes)
            self.log_p_second_best = empty(0)
        elif self.full_step:
            self.old_clusters = self.clusters
            self.clusters = -ones(num_spikes, dtype=int)
            self.clusters_second_best = -ones(num_spikes, dtype=int)
            if hasattr(self, 'log_p_best'):
                self.old_log_p_best = self.log_p_best
            self.log_p_best = inf*ones(num_spikes)
            self.log_p_second_best = inf*ones(num_spikes)
        else:
            self.old_clusters = self.clusters.copy()

        num_skipped = 0

        if self.full_step and self.use_noise_cluster and not only_evaluate_current_clusters:
            # start with cluster 0 - uniform distribution over space
            # because we have normalized all dims to 0...1, density will be 1.
            self.clusters[:] = self.noise_cluster
            self.log_p_best[:] = -log(noise_weight)

        if only_evaluate_current_clusters:
            self.quick_step_candidates = dict()
            for cluster in range(num_clusters):
                self.quick_step_candidates[cluster] = self.get_spikes_in_cluster(cluster)
            self.collect_candidates = False
        elif self.full_step:
            self.quick_step_candidates = dict()
            self.collect_candidates = True
        else:
            self.collect_candidates = False
        
        clusters_to_kill = []
        
        for cluster in range(num_clusters):
            # Compute the sum of fmasks
            if cluster<self.num_special_clusters:
                cluster_mask_sum = full(num_features, -1.0) # ensure that special clusters are masked
            else:
                cluster_mask_sum = zeros(num_features)
                accumulate_cluster_mask_sum(self, cluster_mask_sum, self.get_spikes_in_cluster(cluster))

            # Compute the masked and unmasked sets
            unmasked, = (cluster_mask_sum>=self.points_for_cluster_mask).nonzero()
            masked, = (cluster_mask_sum<self.points_for_cluster_mask).nonzero()
            unmasked = array(unmasked, dtype=int)
            masked = array(masked, dtype=int)
            cov = BlockPlusDiagonalMatrix(masked, unmasked)

            ########### M step ########################################################
        
            # Normalize by total number of points to give class weight
            if cluster==self.noise_cluster:
                weight = noise_weight
            elif cluster==self.mua_cluster:
                weight = (num_cluster_members[self.mua_cluster]+self.mua_point)/denom
            else:
                weight = (num_cluster_members[cluster]+self.prior_point)/denom
        
            # Compute means for each cluster
            # Note that we do this densely at the moment, might want to switch
            # that to a sparse structure later
            cluster_mean = compute_cluster_mean(self, cluster)        
            # Compute covariance matrices
            compute_covariance_matrix(self, cluster, cluster_mean, cov)
            
            ########### EC steps ######################################################
            
            if cluster<self.num_special_clusters:
                continue   
            try:
                chol = cov.cholesky()
            except LinAlgError:
                self.log('warning', 'Linear algebra error on cluster '+str(cluster))
                clusters_to_kill.append(cluster) # todo: we don't actually do anything with this...
                continue

            # LogRootDet is given by log of product of diagonal elements
            log_root_det = sum(log(chol.diagonal))+sum(log(chol.block.diagonal()))

            # compute diagonal of inverse of cov matrix
            inv_cov_diag = zeros(num_features)
            basis_vector = zeros(num_features)
            for i in range(num_features):
                basis_vector[i] = 1.0
                root = chol.trisolve(basis_vector)
                inv_cov_diag[i] = sum(root**2)
                basis_vector[i] = 0.0

            self.run_callbacks('e_step_before_main_loop', cholesky=chol, cluster=cluster,
                               inv_cov_diag=inv_cov_diag)
                
            compute_log_p_and_assign(self, cluster, weight, inv_cov_diag, log_root_det, chol,
                                     cluster_mean, only_evaluate_current_clusters, self.num_cpus)
            
            self.run_callbacks('e_step_after_main_loop')

        # we've reassigned clusters so we need to recompute the partitions, but we don't want to
        # reindex yet because we may reassign points to different clusters and we need the original
        # cluster numbers for that
        self.partition_clusters()

    @add_slots
    def compute_cluster_penalties(self, clusters=None):
        cluster_penalties = compute_penalties(self, clusters)
        if clusters is None:
            self.cluster_penalty = cluster_penalties
        return cluster_penalties

    @add_slots
    def consider_deletion(self):
        num_cluster_members = self.num_cluster_members
        num_clusters = self.num_clusters_alive
        if num_clusters<=self.num_special_clusters:
            self.log('info', 'Not enough clusters to try deletion')
            return
        sic = self.spikes_in_cluster
        sico = self.spikes_in_cluster_offset
        log_p_best = self.log_p_best
        log_p_second_best = self.log_p_second_best

        deletion_loss = zeros(num_clusters)
        I = arange(self.num_spikes)
        add.at(deletion_loss, self.clusters, log_p_second_best-log_p_best)

        score, score_raw, score_penalty = self.compute_score()
        candidate_cluster = -1
        improvement = -inf
        for cluster in range(self.num_special_clusters, num_clusters):
            new_clusters = self.clusters.copy()
            # reassign points
            cursic = sic[sico[cluster]:sico[cluster+1]]
            new_clusters[cursic] = self.clusters_second_best[cursic]
            # compute penalties if we reassigned this
            penalties = self.compute_cluster_penalties(new_clusters)
            new_score = score_raw+deletion_loss[cluster]+sum(penalties)
            cur_improvement = score-new_score # we want improvement to be a positive value
            if cur_improvement>improvement:
                improvement = cur_improvement
                candidate_cluster = cluster

        if improvement>0:
            # delete this cluster
            num_points_in_candidate = sico[candidate_cluster+1]-sico[candidate_cluster]
            self.log('info', 'Deleting cluster {cluster} ({numpoints} points): improves score '
                             'by {improvement}'.format(cluster=candidate_cluster,
                                                       numpoints=num_points_in_candidate,
                                                       improvement=improvement))
            # reassign points
            cursic = sic[sico[candidate_cluster]:sico[candidate_cluster+1]]
            self.clusters[cursic] = self.clusters_second_best[cursic]
            self.log_p_best[cursic] = self.log_p_second_best[cursic]
            # at this point we have invalidated the partitions, so to make sure we don't miss
            # something, we wipe them out here
            self.partition_clusters()
            self.compute_cluster_penalties() # and recompute the penalties
            # we've also invalidated the second best log_p and clusters
            self.log_p_second_best = None
            self.clusters_second_best = None
            # and we will need to do a full step next time
            self.force_next_step_full = True

    @add_slots
    def compute_score(self):
        penalty = sum(self.cluster_penalty)
        raw = sum(self.log_p_best)
        score = raw+penalty
        self.log('debug', 'compute_score: raw %f + penalty %f = %f' % (raw, penalty, score))
        return score, raw, penalty

    @property
    def num_spikes(self):
        return self.data.num_spikes

    @property
    def num_features(self):
        return self.data.num_features#

    @property
    def num_clusters_alive(self):
        return len(self.num_cluster_members)

    @add_slots
    def reindex_clusters(self):
        '''
        Remove any clusters with 0 members (except for clusters 0 and 1),
        and recompute the list of spikes in each cluster. After this function is
        run, you can use the attributes:

        - num_cluster_members (of length the number of clusters)
        - spikes_in_cluster, spikes_in_cluster_offset

        spikes_in_cluster[spikes_in_cluster_offset[c]:spikes_in_cluster_offset[c+1]] will be in the indices
        of all the spikes in cluster c.
        '''
        num_cluster_members = array(bincount(self.clusters), dtype=int)
        I = num_cluster_members>0
        I[0:self.num_special_clusters] = True # we keep special clusters
        remapping = hstack((0, cumsum(I)))[:-1]
        self.clusters = remapping[self.clusters]
        total_clusters = sum(I)
        if hasattr(self, '_total_clusters') and total_clusters<self._total_clusters:
            self.force_next_step_full = True
            if hasattr(self, 'clusters_second_best'):
                del self.clusters_second_best
        self._total_clusters = total_clusters
        self.partition_clusters()

    def partition_clusters(self, clusters=None):
        if clusters is None:
            clusters = self.clusters
            assign_to_self = True
        else:
            assign_to_self = False
        try:
            if self.num_special_clusters>0:
                num_cluster_members = array(bincount(clusters, minlength=self.num_special_clusters), dtype=int)
            else:
                num_cluster_members = array(bincount(clusters), dtype=int)
        except ValueError:
            raise PartitionError
        I = array(argsort(clusters), dtype=int)
        y = clusters[I]
        n = amax(y)
        if n<self.num_special_clusters-1:
            n = self.num_special_clusters-1
        n += 2
        J = searchsorted(y, arange(n))
        if assign_to_self:
            self.num_cluster_members = num_cluster_members
            self.spikes_in_cluster = I
            self.spikes_in_cluster_offset = J
        else:
            return I, J, num_cluster_members

    def invalidate_partitions(self):
        self.num_cluster_members = None
        self.spikes_in_cluster = None
        self.spikes_in_cluster_offset = None

    def get_spikes_in_cluster(self, cluster):
        sic = self.spikes_in_cluster
        sico = self.spikes_in_cluster_offset
        return sic[sico[cluster]:sico[cluster+1]]
        
    @add_slots
    def try_splits(self):
        did_split = False
        num_clusters = self.num_clusters_alive

        self.log('info', 'Trying to split clusters')

        score_ref = None

        self.reindex_clusters()

        for cluster in range(self.num_special_clusters, num_clusters):
            if num_clusters>=self.max_possible_clusters:
                self.log('info', 'No more splitting, already at maximum number of '
                                 'clusters: %d' % self.max_possible_clusters)
                return did_split

            spikes_in_cluster = self.get_spikes_in_cluster(cluster)
            if len(spikes_in_cluster)==0:
                continue

            with section(self, 'split_candidate'):
                if self.max_split_iterations is not None:
                    max_iter = self.max_split_iterations
                else:
                    max_iter = self.max_iterations

                K2 = self.subset(spikes_in_cluster, name='split_candidate',
                                 use_noise_cluster=False, use_mua_cluster=False,
                                 max_iterations=max_iter,
                                 map_log_to_debug=True,
                                 )
                # at this point in C++ code we look for an unused cluster, but here we can just
                # use num_clusters+1
                self.log('debug', 'Trying to split cluster %d containing '
                                  '%d points' % (cluster, len(spikes_in_cluster)))
                # initialise with current clusters, do not allow creation of new clusters
                K2.max_possible_clusters = 1
                clusters = full(len(spikes_in_cluster), 0, dtype=int)
                try:
                    unsplit_score = K2.cluster_from(clusters, recurse=False)
                except PartitionError:
                    self.log('error', 'Partitioning error on split, K2.clusters = %s' % K2.clusters)
                    continue
                self.run_callbacks('split_k2_1', cluster=cluster, K2=K2,
                                   unsplit_score=unsplit_score)
                # initialise randomly, allow for one additional cluster
                K2.max_possible_clusters = 2
                clusters = randint(0, 2, size=len(spikes_in_cluster))
                if amax(clusters)!=1:
                    continue

                if self.fast_split:
                    score_target = unsplit_score
                else:
                    score_target = -inf

                try:
                    split_score = K2.cluster_from(clusters, recurse=False,
                                                  score_target=score_target)
                except PartitionError:
                    self.log('error', 'Partitioning error on split, K2.clusters = %s' % K2.clusters)
                    continue
                self.run_callbacks('split_k2_2', cluster=cluster, K2=K2, split_score=split_score,
                                   unsplit_score=unsplit_score)

                if K2.num_clusters_alive==0:
                    self.log('error', 'No clusters alive in K2')
                    continue

                if split_score>=unsplit_score:
                    self.log('debug', 'Score after (%f) splitting worse than before (%f), '
                                      'so not splitting' % (split_score, unsplit_score))
                    continue

            if self.always_split_bimodal:
                self.log('debug', 'Always splitting bimodal clusters, so splitting cluster '
                                  '%d into %d' % (cluster, num_clusters))
                clusters = self.clusters.copy()
                I1 = (K2.clusters==1)
                clusters[spikes_in_cluster[I1]] = num_clusters # next available cluster
                did_split = True
                self.clusters = clusters
                self.reindex_clusters()
                num_clusters = self.num_clusters_alive
                continue

            with section(self, 'split_evaluation'):
                # will splitting improve the score in the whole data set?
                K3 = self.copy(name='split_evaluation', map_log_to_debug=True)
                clusters = self.clusters.copy()

                if score_ref is None:
                    K3.initialise_clusters(clusters)
                    K3.prepare_for_iterate()
                    K3.MEC_steps(only_evaluate_current_clusters=True)
                    K3.compute_cluster_penalties()
                    score_ref, _, _ = K3.compute_score()

                I1 = (K2.clusters==1)
                clusters[spikes_in_cluster[I1]] = num_clusters # next available cluster

                K3.initialise_clusters(clusters)
                K3.prepare_for_iterate()
                K3.MEC_steps(only_evaluate_current_clusters=True)
                K3.compute_cluster_penalties()
                score_new, _, _ = K3.compute_score()

            if score_new<score_ref:
                self.log('debug', 'Score improved after splitting, so splitting cluster '
                                  '%d into %d' % (cluster, num_clusters))
                did_split = True
                self.clusters = K3.clusters.copy()
                self.reindex_clusters()
                num_clusters = self.num_clusters_alive
                score_ref = score_new
            else:
                self.log('debug', 'Score got worse after splitting')

        # if we split, should make the next step full
        if did_split:
            self.force_next_step_full = True
            self.log('info', 'Split into %d clusters' % num_clusters)

        return did_split



from numpy import *
from klustakwik2 import *
from numpy.testing import assert_raises, assert_array_almost_equal, assert_array_equal
from nose import with_setup
from nose.tools import nottest

from tempfile import mkdtemp
import os
from six.moves import range

@nottest
def generate_simple_test_raw_data():
    # Data as files
    dirname = mkdtemp()
    open(os.path.join(dirname, 'kktest.fet.1'), 'wt').write('''
5
1 3 5 7 11
6 7 8 9 10
11 12 13 14 15
16 17 18 19 20
'''.strip())
    open(os.path.join(dirname, 'kktest.fmask.1'), 'wt').write('''
5
1 0.5 0 0 0
0 1 1 0 0
0 1 1 0 0
0 0 0 1 1
'''.strip())

    # Data as arrays
    fet = array([[1,3,5,7,11],
                 [6,7,8,9,10],
                 [11,12,13,14,15],
                 [16,17,18,19,20]], dtype=float)
    fmask = array([[1,0.5,0,0,0],
                   [0,1,1,0,0],
                   [0,1,1,0,0],
                   [0,0,0,1,1]], dtype=float)

    # Normalisation to [0, 1]
    fet = (fet-amin(fet, axis=0))/(amax(fet, axis=0)-amin(fet, axis=0))

    nanmasked_fet = fet.copy()
    nanmasked_fet[fmask>0] = nan

    # Correct computation of the corrected data and correction term
    x = fet
    w = fmask
    nu = nanmean(nanmasked_fet, axis=0)[newaxis, :]
    sigma2 = nanvar(nanmasked_fet, axis=0)[newaxis, :]
    y = w*x+(1-w)*nu
    z = w*x*x+(1-w)*(nu*nu+sigma2)
    correction_terms = z-y*y
    features = y

    return (load_fet_fmask_to_raw(os.path.join(dirname, 'kktest'), 1),
            fet, fmask, features, correction_terms)


@nottest
def generate_simple_test_data():
    data, fet, fmask, features, correction_terms = generate_simple_test_raw_data()
    return data.to_sparse_data()


def test_load_fet_fmask():

    ### PART 1: Check that loading to RawSparseData is correct
    data, fet, fmask, features, correction_terms = generate_simple_test_raw_data()

    nanmasked_fet = fet.copy()
    nanmasked_fet[fmask>0] = nan
    assert_array_almost_equal(data.noise_mean, nanmean(nanmasked_fet, axis=0))
    assert_array_almost_equal(data.noise_variance, nanvar(nanmasked_fet, axis=0))
    assert amin(data.features)==0
    assert amax(data.features)==1
    assert len(data.offsets)==5
    loaded_fet = zeros_like(fet)
    for i in range(4):
        data_u = data.unmasked[data.offsets[i]:data.offsets[i+1]]
        true_u, = fmask[i, :].nonzero()
        assert_array_equal(data_u, true_u)
        data_f = data.features[data.offsets[i]:data.offsets[i+1]]
        true_f = fet[i, data_u]
        assert_array_equal(data_f, true_f)
        data_m = data.masks[data.offsets[i]:data.offsets[i+1]]
        true_m = fmask[i, data_u]
        assert_array_equal(data_m, true_m)

    ### PART 2: Check that converting to SparseData is correct
    data = data.to_sparse_data() # compute unique masks and apply correction terms to data

    assert data.num_spikes==4
    assert data.num_features==5
    assert data.num_masks==3

    for i in range(4):
        data_u = data.unmasked[data.unmasked_start[i]:data.unmasked_end[i]]
        true_u, = fmask[i, :].nonzero()
        assert_array_equal(data_u, true_u)
        data_f = data.features[data.values_start[i]:data.values_end[i]]
        true_f = features[i, data_u]
        assert_array_almost_equal(data_f, true_f)
        data_c = data.correction_terms[data.values_start[i]:data.values_end[i]]
        true_c = correction_terms[i, data_u]
        assert_array_almost_equal(data_c, true_c)
        data_m = data.masks[data.values_start[i]:data.values_end[i]]
        true_m = fmask[i, data_u]
        assert_array_almost_equal(data_m, true_m)


def test_subset_features():
    raw_data, fet, fmask, features, correction_terms = generate_simple_test_raw_data()
    data = raw_data.to_sparse_data()
    for I, sp in [([0], [0]),
                  ([3, 4], [3]),
                  ([1, 2], [0, 1, 2]),
                  ([1, 3, 4], [0, 1, 2, 3]),
                  ]:
        I = array(I, dtype=int)
        subdata, spikes = data.subset_features(I)
        assert_array_equal(spikes, sp)
        U = lambda p: subdata.unmasked[subdata.unmasked_start[p]:subdata.unmasked_end[p]]
        F = lambda p: subdata.features[subdata.values_start[p]:subdata.values_end[p]]
        M = lambda p: subdata.masks[subdata.values_start[p]:subdata.values_end[p]]
        CT = lambda p: subdata.correction_terms[subdata.values_start[p]:subdata.values_end[p]]
        for p in range(len(spikes)):
            assert_array_equal(U(p), (fmask[spikes[p], I]>0).nonzero()[0])
            assert_array_equal(F(p), features[spikes[p], I][fmask[spikes[p], I]>0])
            assert_array_equal(M(p), fmask[spikes[p], I][fmask[spikes[p], I]>0])
            assert_array_equal(CT(p), correction_terms[spikes[p], I][fmask[spikes[p], I]>0])

if __name__=='__main__':
    test_load_fet_fmask()
    test_subset_features()




from numpy import *
import time

from .data import RawSparseData
from .logger import log_message
from six.moves import zip

__all__ = ['load_fet_fmask_to_raw', 'save_clu', 'load_clu', 'SaveCluEvery']


def load_fet_fmask_to_raw(fname, shank, use_features=None, drop_last_n_features=0):
    if use_features is None and drop_last_n_features>0:
        use_features = slice(None, -drop_last_n_features)
    else:
        use_features = slice(None)
    fet_fname = fname+'.fet.'+str(shank)
    fmask_fname = fname+'.fmask.'+str(shank)
    # read files
    fet_file = open(fet_fname, 'r')
    fmask_file = open(fmask_fname, 'r')
    # read first line of fmask file
    num_features = int(fmask_file.readline())
    features_to_use = arange(num_features)[use_features]
    num_features = len(features_to_use)
    # Stage 1: read min/max of fet values for normalisation
    # and count total number of unmasked features
    fet_file.readline() # skip first line (num channels)
    # we normalise channel-by-channel
    vmin = ones(num_features)*inf
    vmax = ones(num_features)*-inf
    total_unmasked_features = 0
    num_spikes = 0
    for fetline, fmaskline in zip(fet_file, fmask_file):
        vals = fromstring(fetline, dtype=float, sep=' ')[use_features]
        fmaskvals = fromstring(fmaskline, dtype=float, sep=' ')[use_features]
        inds, = (fmaskvals>0).nonzero()
        total_unmasked_features += len(inds)
        vmin = minimum(vals, vmin)
        vmax = maximum(vals, vmax)
        num_spikes += 1
    # Stage 2: read data line by line, normalising
    fet_file.close()
    fet_file = open(fet_fname, 'r')
    fet_file.readline()
    fmask_file.close()
    fmask_file = open(fmask_fname, 'r')
    fmask_file.readline()
    vdiff = vmax-vmin
    vdiff[vdiff==0] = 1
    fetsum = zeros(num_features)
    fet2sum = zeros(num_features)
    nsum = zeros(num_features)
    all_features = zeros(total_unmasked_features)
    all_fmasks = zeros(total_unmasked_features)
    all_unmasked = zeros(total_unmasked_features, dtype=int)
    offsets = zeros(num_spikes+1, dtype=int)
    curoff = 0
    for i, (fetline, fmaskline) in enumerate(zip(fet_file, fmask_file)):
        fetvals = (fromstring(fetline, dtype=float, sep=' ')[use_features]-vmin)/vdiff
        fmaskvals = fromstring(fmaskline, dtype=float, sep=' ')[use_features]
        inds, = (fmaskvals>0).nonzero()
        masked_inds, = (fmaskvals==0).nonzero()
        all_features[curoff:curoff+len(inds)] = fetvals[inds]
        all_fmasks[curoff:curoff+len(inds)] = fmaskvals[inds]
        all_unmasked[curoff:curoff+len(inds)] = inds
        offsets[i] = curoff
        curoff += len(inds)
        fetsum[masked_inds] += fetvals[masked_inds]
        fet2sum[masked_inds] += fetvals[masked_inds]**2
        nsum[masked_inds] += 1
    offsets[-1] = curoff

    nsum[nsum==0] = 1
    noise_mean = fetsum/nsum
    noise_variance = fet2sum/nsum-noise_mean**2

    return RawSparseData(noise_mean, noise_variance, all_features, all_fmasks, all_unmasked, offsets)


def save_clu(kk, fname, shank):
    savetxt(fname+'.clu.'+str(shank), kk.clusters+1, '%d', header=str(amax(kk.clusters)), comments='')


def load_clu(fname):
    return loadtxt(fname, skiprows=1, dtype=int)-1


class SaveCluEvery(object):
    '''
    Callback to save the clu file every fixed number of iterations.

    fname can be a simple string, or a formattable string that takes the kk object as an argument,
    so that you can write, e.g. fname='testdata.{kk.name}.{kk.current_iteration}'.

    every is the number of minutes between saves

    save_all=True will append .iterN for N the iteration number to the filename, and therefore save
    each clu file produced.
    '''
    def __init__(self, fname, shank, every=1, save_all=False):
        self.fname = fname
        self.shank = shank
        self.every = every*60
        self.save_all = save_all
        self.t_next = time.time()+self.every

    def __call__(self, kk):
        t_cur = time.time()
        if kk.name=='' and t_cur>self.t_next and not kk.is_subset:
            shank = str(self.shank)
            if self.save_all:
                shank = str(shank)+'.iter'+str(kk.current_iteration)
            fname = self.fname.format(kk=kk)
            log_message('info', 'Saving clu to file '+fname+'.clu.'+shank)
            save_clu(kk, fname, shank)
            self.t_next = time.time()+self.every




from numpy import *
import time

from .precomputations import (reduce_masks, compute_correction_terms_and_replace_data,
                              compute_float_num_unmasked, reduce_masks_from_arrays,
                              compute_float_num_unmasked_from_arrays)
from .logger import log_message
from six.moves import range

__all__ = ['RawSparseData', 'SparseData',
           ]


class RawSparseData(object):
    '''
    Contains raw, sparse data used for clustering
    
    This object is only used as an intermediate step, not directly by the
    other algorithms. It should be converted into a SparseData object.
    
    Note that for .fet/.fmask data there already has to be some non-trivial
    processing to get this far (to sparsify the data and extract the
    noise mean and variance). However, in the future this step may have been
    computed by SpikeDetekt which will output sparse data.
    
    Also note that the indices can be a smaller array than the features and
    fmasks because there may be a much smaller unique set.
    TODO: method to carry this out
    
    Consists of the following:

    - noise_mean, noise_variance: arrays of length the number of features
    - features: array of feature values
    - masks: array of mask values
    - unmasked: array of unmasked indices
    - offset: array of offset indices (of length the number of spikes + 1)
    
    This is a sparse array such that the feature vector for spike i has
    non-default values at indices unmasked[offset[i]:offset[i+1]]
    with values features[offset[i]:offset[i+1]]. masks has the same structure.
    
    These arrays have to be normalised in the following fashion:
    - For each channel, the features have to be first normalised between 0 and 1 (the linear
      transformation is different for each channel).
    - The noise mean and noise variance are computed on this normalised data.
    - Unmasked data is where fmask>0, masked data is where fmask==0
    - Only masked data is used for computing the noise mean and noise variance.
    '''
    def __init__(self,
                 noise_mean, noise_variance,
                 features, masks, unmasked,
                 offsets,
                 ):
        # Raw data
        self.noise_mean = noise_mean
        self.noise_variance = noise_variance
        self.features = features
        self.masks = masks
        self.unmasked = unmasked
        self.offsets = offsets
        
    def to_sparse_data(self):
        values_start = self.offsets[:-1]
        values_end = self.offsets[1:]
        log_message('debug', 'Starting compute_correction_terms_and_replace_data', name='data')
        features, correction_terms = compute_correction_terms_and_replace_data(self)
        log_message('debug', 'Finished compute_correction_terms_and_replace_data', name='data')
        log_message('debug', 'Starting sort_masks', name='data')
        unmasked, unmasked_start, unmasked_end = reduce_masks(self)
        log_message('debug', 'Finished sort_masks', name='data')
        float_num_unmasked = compute_float_num_unmasked(self)
        return SparseData(self.noise_mean, self.noise_variance,
                          features, self.masks,
                          values_start, values_end,
                          unmasked,
                          unmasked_start, unmasked_end,
                          correction_terms,
                          float_num_unmasked,
                          )


class SparseData(object):
    '''
    Notes:
    - Assumes that the spikes are in sorted mask order, can use unmasked_start
      as a proxy for the identity of the mask
    '''
    def __init__(self,
                 noise_mean, noise_variance,
                 features, masks,
                 values_start, values_end,
                 unmasked,
                 unmasked_start, unmasked_end,
                 correction_terms,
                 float_num_unmasked,
                 ):
        # Data arrays
        self.noise_mean = noise_mean
        self.noise_variance = noise_variance
        self.features = features
        self.masks = masks
        self.values_start = values_start
        self.values_end = values_end
        self.unmasked = unmasked
        self.unmasked_start = unmasked_start
        self.unmasked_end = unmasked_end
        self.correction_terms = correction_terms
        self.float_num_unmasked = float_num_unmasked
        # Derived data
        self.num_spikes = len(self.values_start)
        self.num_features = len(self.noise_mean)
        
        self.num_masks = len(unique(self.unmasked_start))
        
    def subset(self, spikes):
        return SparseData(self.noise_mean, self.noise_variance,
                          self.features, self.masks,
                          self.values_start[spikes], self.values_end[spikes],
                          self.unmasked,
                          self.unmasked_start[spikes], self.unmasked_end[spikes],
                          self.correction_terms,
                          self.float_num_unmasked[spikes],
                          )

    def subset_features(self, feature_indices):
        '''
        Returns a pair data, spikes where spikes are those spikes that are not masked on every channel.
        '''
        feature_indices = array(feature_indices, dtype=int)
        # do the easy parts
        noise_mean = self.noise_mean[feature_indices]
        noise_variance = self.noise_variance[feature_indices]

        new_feature_indices = full(self.num_features, -1, dtype=self.unmasked.dtype)
        new_feature_indices[feature_indices] = arange(len(feature_indices))

        # inefficient method, loop through all spikes
        spikes = []
        features = []
        masks = []
        offsets = [0]
        total_features = 0
        unmasked = []
        correction_terms = []
        for p in range(self.num_spikes):
            U = self.unmasked[self.unmasked_start[p]:self.unmasked_end[p]]
            U_in = in1d(U, feature_indices)
            if sum(U_in):
                F = self.features[self.values_start[p]:self.values_end[p]]
                M = self.masks[self.values_start[p]:self.values_end[p]]
                CT = self.correction_terms[self.values_start[p]:self.values_end[p]]
                U2, = U_in.nonzero()
                F2 = F[U2]
                M2 = M[U2]
                CT2 = CT[U2]
                spikes.append(p)
                features.append(F2)
                masks.append(M2)
                correction_terms.append(CT2)
                unmasked.append(new_feature_indices[U[U_in]])
                total_features += len(F2)
                offsets.append(total_features)
        features = array(hstack(features), dtype=self.features.dtype)
        masks = array(hstack(masks), dtype=self.masks.dtype)
        correction_terms = array(hstack(correction_terms), dtype=self.correction_terms.dtype)
        unmasked = array(hstack(unmasked), dtype=self.unmasked.dtype)
        spikes = array(spikes, dtype=int)
        offsets = array(offsets, dtype=int)
        values_start = offsets[:-1]
        values_end = offsets[1:]

        assert amin(unmasked)>=0

        unmasked, unmasked_start, unmasked_end = reduce_masks_from_arrays(values_start, values_end, unmasked)
        float_num_unmasked = compute_float_num_unmasked_from_arrays(masks, offsets)

        data = SparseData(noise_mean, noise_variance,
                          features, masks,
                          values_start, values_end,
                          unmasked, unmasked_start, unmasked_end,
                          correction_terms,
                          float_num_unmasked,
                          )

        return data, spikes



import numpy
import numba
from six.moves import range

def compute_cluster_means(kk):
    num_clusters = len(kk.num_cluster_members)
    num_features = kk.num_features
    data = kk.data

    cluster_mean = numpy.zeros((num_clusters, num_features))
    num_added = numpy.zeros((num_clusters, num_features), dtype=int)

    _compute_cluster_means(kk.clusters,
                           data.unmasked, data.unmasked_start, data.unmasked_end,
                           data.features, data.values_start, data.values_end,
                           cluster_mean, num_added,
                           kk.mua_point, kk.prior_point,
                           data.noise_mean, kk.num_cluster_members)

    return cluster_mean

@numba.jit(nopython=True)
def _compute_cluster_means(
            clusters,
            unmasked, ustart, uend,
            features, vstart, vend,
            cluster_mean, num_added,
            mua_point, prior_point,
            noise_mean, num_cluster_members,
            ):

    num_clusters = len(num_cluster_members)
    num_features = cluster_mean.shape[1]

    for p in range(len(clusters)):
        c = clusters[p]
        num_unmasked = uend[p]-ustart[p]
        for i in range(num_unmasked):
            j = unmasked[ustart[p]+i]
            k = vstart[p]+i
            cluster_mean[c, j] += features[k]
            num_added[c, j] += 1

    for cluster in range(num_clusters):
        if num_cluster_members[cluster]==0:
            continue
        prior = 0
        if cluster==1:
            prior = mua_point
        elif cluster>=2:
            prior = prior_point
        for i in range(num_features):
            cluster_mean[cluster, i] += noise_mean[i]*(num_cluster_members[cluster]-num_added[cluster, i])
            cluster_mean[cluster, i] += prior*noise_mean[i]
            cluster_mean[cluster, i] /= num_cluster_members[cluster]+prior



from numpy import *
from .masks_cy import *
from random import shuffle

__all__ = ['sort_masks', 'mask_start_clusters']

def sort_masks(raw_data):
    # step 1: sort into lexicographical order of masks
    O = raw_data.offsets
    I = raw_data.unmasked
    start = zeros(len(O)-1, dtype=int)
    end = zeros(len(O)-1, dtype=int)
    new_indices = zeros_like(I)

    num_new_indices = do_sort_masks(I, O, start, end, new_indices)
    
    new_indices = array(new_indices[:num_new_indices], copy=True)

    return None, new_indices, start, end


def mask_start_clusters(data, num_clusters):
    clusters = full(data.num_spikes, -1, dtype=int) # start with -1 to make sure we don't miss any
    num_features = data.num_features

    # go through the spikes in random order so as not to bias the selection of masks
    allspikes = arange(data.num_spikes)
    shuffle(allspikes)
    
    do_mask_starts(clusters,
                   data.unmasked, data.unmasked_start, data.unmasked_end,
                   allspikes, num_features, num_clusters)

    return clusters



from numpy import *
from klustakwik2 import *
from numpy.testing import assert_raises, assert_array_almost_equal, assert_array_equal
from nose import with_setup
from nose.tools import nottest
from numpy.random import randint, rand
from six.moves import range

@nottest
def generate_multimask_test_data(num_masks, num_points, num_features):
    # step 1: generate masks
    found = set()
    all_masks = []
    for i in range(num_masks):
        while True:
            u, = randint(2, size=num_features).nonzero()
            h = u.tostring()
            if h not in found:
                found.add(h)
                all_masks.append(u)
                break
    assert len(all_masks)==num_masks
    # step 2: generate data
    fet = []
    fmask = []
    offsets = []
    unmasked = []
    n = 0
    offsets = [n]
    for i in range(num_points):
        u = all_masks[randint(num_masks)]
        fet.append(rand(len(u)))
        fmask.append(0.5+0.5*rand(len(u)))
        unmasked.append(u)
        n += len(u)
        offsets.append(n)
    assert len(offsets)==num_points+1
    fet = hstack(fet)
    fmask = hstack(fmask)
    unmasked = array(hstack(unmasked), dtype=int)
    offsets = array(offsets)
    return RawSparseData(full(num_features, 0.5), full(num_features, 1./12), # mean/var of rand()
                         fet, fmask, unmasked, offsets).to_sparse_data()


def test_mask_starts():
    data = generate_multimask_test_data(100, 1000, 10)

    assert data.num_spikes==1000
    assert data.num_features==10
    assert 90<=data.num_masks<=110 # a bit of leeway because it's random

    # too few clusters raises an error
    assert_raises(ValueError, mask_starts, data, 0, 2)
    assert_raises(ValueError, mask_starts, data, 2, 2)

    # too many clusters logs a warning and gives us as many as possible
    clusters = mask_starts(data, 150, 2)
    assert amin(clusters)==2
    # 102 because we aim for 150 but only have 100 masks, which is 102 clusters (0,1=noise,mua)
    assert 95<amax(clusters)<102

    # half the maximum number of clusters should be ok and give us something to test
    clusters = mask_starts(data, 50, 2)
    assert amin(clusters)==2
    assert 45<amax(clusters)<50 # a bit of leeway because it's random

    # Basic sanity check on the generated clusters
    for p in range(len(clusters)):
        cluster = clusters[p]
        unmasked = data.unmasked[data.unmasked_start[p]:data.unmasked_end[p]]
        for p2 in range(len(clusters)):
            cluster2 = clusters[p2]
            #unmasked2 = data.unmasked[data.unmasked_start[p2]:data.unmasked_end[p2]]
            # this says that if the masks are the same they ought to be assigned to the same
            # cluster
            if data.unmasked_start[p]==data.unmasked_start[p2]:
                assert cluster2==cluster


if __name__=='__main__':
    test_mask_starts()



'''
Linear algebra routines
'''

from numpy import *
from numpy.linalg import cholesky, LinAlgError
from scipy.linalg import cho_factor, cho_solve, cholesky, solve_triangular

__all__ = ['BlockPlusDiagonalMatrix']

class BlockPlusDiagonalMatrix(object):
    def __init__(self, masked, unmasked, block=None, diagonal=None):
        self.masked = masked
        self.unmasked = unmasked
        self.num_masked = len(masked)
        self.num_unmasked = len(unmasked)
        if block is None:
            block = zeros((self.num_unmasked, self.num_unmasked))
        if diagonal is None:
            diagonal = zeros(self.num_masked)
        self.block = block
        self.diagonal = diagonal
    
    def new_with_same_masks(self, block=None, diagonal=None):
        if block is not None:
            block = block.copy()
        if diagonal is not None:
            diagonal = diagonal.copy()
        return BlockPlusDiagonalMatrix(self.masked, self.unmasked, block=block, diagonal=diagonal)

    def cholesky(self):
        M_chol = self.new_with_same_masks()
        lower = None
        if self.block.size:
            block = cholesky(self.block, lower=True)
        else:
            block = zeros_like(self.block)
        if (self.diagonal<=0).any():
            raise LinAlgError
        diagonal = sqrt(self.diagonal)
        return self.new_with_same_masks(block, diagonal)
    
    def trisolve(self, x):
        out = zeros(len(x))
        if len(self.unmasked):
            out[self.unmasked] = solve_triangular(self.block, x[self.unmasked], lower=True)
        out[self.masked] = x[self.masked]/self.diagonal
        return out



from numpy import *
from klustakwik2 import *
from numpy.testing import assert_raises, assert_array_almost_equal, assert_array_equal
from nose import with_setup
from nose.tools import nottest
from numpy.random import randint, rand
from copy import deepcopy
from six.moves import range
from klustakwik2.linear_algebra import BlockPlusDiagonalMatrix

# we use the version that is used in klustakwik2 rather than separately testing the numba/cython
# versions
from klustakwik2.clustering import (accumulate_cluster_mask_sum, compute_cluster_mean,
                                    compute_covariance_matrix)

from .test_compute_cluster_masks import generate_simple_test_kk

def test_compute_covariance_matrix():
    kk = generate_simple_test_kk(points_for_cluster_mask=0.5)

    num_cluster_members = kk.num_cluster_members
    num_clusters = kk.num_clusters_alive
    num_features = kk.num_features

    cluster_mask_sum = zeros((num_clusters, num_features))
    for cluster in range(kk.num_special_clusters, num_clusters):
        accumulate_cluster_mask_sum(kk, cluster_mask_sum[cluster, :], kk.get_spikes_in_cluster(cluster))
    cluster_mask_sum[:kk.num_special_clusters, :] = -1 # ensure that special clusters are masked

    cov_matrices = []

    for cluster in range(1, num_clusters):
        cluster_mean = compute_cluster_mean(kk, cluster)
        curmask = cluster_mask_sum[cluster, :]
        unmasked, = (curmask>=kk.points_for_cluster_mask).nonzero()
        masked, = (curmask<kk.points_for_cluster_mask).nonzero()
        unmasked = array(unmasked, dtype=int)
        masked = array(masked, dtype=int)
        cov = BlockPlusDiagonalMatrix(masked, unmasked)
        compute_covariance_matrix(kk, cluster, cluster_mean, cov)
        f2m = kk.orig_features-cluster_mean[newaxis, :]
        ct = kk.orig_correction_terms
        spikes = kk.get_spikes_in_cluster(cluster)
        block = zeros_like(cov.block)
        unmasked = cov.unmasked
        for spike in spikes:
            if cluster==1:
                for i in range(len(unmasked)):
                    block[i, i] += f2m[spike, unmasked[i]]**2
            else:
                block[:, :] += f2m[spike, unmasked][:, newaxis]*f2m[spike, unmasked][newaxis, :]
            for i in range(len(unmasked)):
                block[i, i] += ct[spike, unmasked[i]]
        if cluster==1:
            point = kk.mua_point
        else:
            point = kk.noise_point
        for i in range(len(cov.unmasked)):
            block[i, i] += point*kk.data.noise_variance[cov.unmasked[i]]
        factor = 1.0/(num_cluster_members[cluster]+point-1)
        block *= factor
        assert_array_almost_equal(block, cov.block)
        cov_matrices.append(deepcopy(cov))

    return cov_matrices


if __name__=='__main__':
    test_compute_covariance_matrix()



from .compute_cluster_masks_cy import doaccum

__all__ = ['accumulate_cluster_mask_sum']

def accumulate_cluster_mask_sum(kk, cluster_mask_sum, spikes):
    data = kk.data
    doaccum(spikes, data.unmasked, data.unmasked_start, data.unmasked_end,
            data.masks, data.values_start, data.values_end, cluster_mask_sum)



from numpy import *
from bisect import bisect_left

__all__ = ['compute_correction_terms_and_replace_data',
           'reduce_masks', 'reduce_masks_from_arrays', 'float_num_unmasked',
           'compute_float_num_unmasked_from_arrays',
           'compute_float_num_unmasked,'
           ]

def compute_correction_terms_and_replace_data(raw_data):
    I = raw_data.unmasked
    x = raw_data.features
    w = raw_data.masks
    nu = raw_data.noise_mean[I]
    sigma2 = raw_data.noise_variance[I]
    y = w*x+(1-w)*nu
    z = w*x*x+(1-w)*(nu*nu+sigma2)
    correction_terms = z-y*y
    features = y
    return features, correction_terms


def reduce_masks_from_arrays(Ostart, Oend, I):
    x = arange(len(Ostart))
    # converting the array to a string allows for a lexicographic compare
    # the details of the comparison are irrelevant as long as it is
    # consistent (for sorting) and never equal if the underlying arrays
    # are unequal
    x = array(sorted(x, key=lambda p: I[Ostart[p]:Oend[p]].tostring()), dtype=int)
    y = empty_like(x)
    y[x] = arange(len(x)) # y is the inverse of x as a permutation
    # step 2: iterate through all indices and add to collection if the
    # indices have changed
    oldstr = None
    new_indices = []
    start = zeros(len(Ostart), dtype=int)
    end = zeros(len(Ostart), dtype=int)
    curstart = 0
    curend = 0
    for i, p in enumerate(x):
        curind = I[Ostart[p]:Oend[p]]
        curstr = curind.tostring()
        if curstr!=oldstr:
            new_indices.append(curind)
            oldstr = curstr
            curstart = curend
            curend += len(curind)
        start[i] = curstart
        end[i] = curend
    # step 3: convert into start, end
    new_indices = hstack(new_indices)
    return new_indices, start[y], end[y]


def reduce_masks(raw_data):
    # step 1: sort into lexicographical order of masks
    start = raw_data.offsets[:-1]
    end = raw_data.offsets[1:]
    I = raw_data.unmasked
    return reduce_masks_from_arrays(start, end, I)


def compute_float_num_unmasked(data):
    M = data.masks
    O = data.offsets
    return compute_float_num_unmasked_from_arrays(M, O)


def compute_float_num_unmasked_from_arrays(M, O):
    # Rather than just doing add.reduceat(data.masks, data.offset[:-1]) which is what we'd like to
    # do, we have to avoid any offsets equal to the length of masks because even though
    # the calculation is correct if you think of it as slices, numpy raises an error if any indices
    # are equal to the length of the array. So we just cut those out and add some zeros at the end.
    O = O[:-1]
    n = bisect_left(O, len(M))-1
    O2 = O[:n]
    U = hstack((add.reduceat(M, O2), zeros(len(O)-n)))
    return U



from numpy import *
from .logger import log_message
from random import shuffle
from numpy.random import randint
from .numerics import mask_start_clusters

__all__ = ['mask_starts']

def mask_starts(data, num_clusters, num_special_clusters):
    log_message('info', 'Using mask starts with %d clusters, data has '
                        '%d unique masks' % (num_clusters, data.num_masks))
    if num_clusters<=num_special_clusters:
        raise ValueError("Number of starting clusters must be at least 3 to accomodate noise and mua clusters.")
    num_clusters -= num_special_clusters # special clusters are not generated, we add them at the end
    if data.num_masks<num_clusters:
        log_message('warning', ('Not enough masks (%d) for specified number of starting '
                                'clusters (%d)') % (data.num_masks, num_clusters+2))
        num_clusters = data.num_masks
         
    return mask_start_clusters(data, num_clusters)+num_special_clusters



'''
Tools for monitoring KK as it runs

MonitoringServer and MonitoringClient are lightly adapted from Brian (briansimulator.org).
'''

try:
    import multiprocessing
    from multiprocessing.connection import Listener, Client
except ImportError:
    multiprocessing = None
import select
import inspect
from six import exec_

__all__ = ['MonitoringServer', 'MonitoringClient']


class MonitoringServer(object):
    '''
    Allows remote control (via IP) of a running KK script

    Initialisation arguments:

    ``server``
        The IP server details, a pair (host, port). If you want to allow control
        only on the one machine (for example, by an IPython shell), leave this
        as ``None`` (which defaults to host='localhost', port=2719). To allow
        remote control, use ('', portnumber).
    ``authkey``
        The authentication key to allow access, change it from 'klustakwik' if you
        are allowing access from outside (otherwise you allow others to run
        arbitrary code on your machine).
    ``global_ns``, ``local_ns``, ``level``
        Namespaces in which incoming commands will be executed or evaluated,
        if you leave them blank it will be the local and global namespace of
        the frame from which this function was called (if level=1, or from
        a higher level if you specify a different level here).

    Once this object has been created, use a :class:`MonitoringClient` to
    issue commands.

    **Example usage**

    Main script code includes a line like this::

        server = MonitoringServer()
        kk.register_callback(server)

    In an IPython shell you can do something like this::

        client = MonitoringClient()
        print client.evaluate('kk.num_clusters_alive')
    '''
    def __init__(self, server=None, authkey='klustakwik',
                 global_ns=None, local_ns=None, level=0):
        if multiprocessing is None:
            raise ImportError('Cannot import the required multiprocessing module.')
        if server is None:
            server = ('localhost', 2719)
        frame = inspect.stack()[level + 1][0]
        ns_global, ns_local = frame.f_globals, frame.f_locals
        if global_ns is None:
            global_ns = frame.f_globals
        if local_ns is None:
            local_ns = frame.f_locals
        self.local_ns = local_ns
        self.global_ns = global_ns
        self.listener = Listener(server, authkey=authkey)
        self.conn = None

    def __call__(self, kk):
        try:
            if self.conn is None:
                # This is kind of a hack. The multiprocessing.Listener class doesn't
                # allow you to tell if an incoming connection has been requested
                # without accepting that connection, which means if nothing attempts
                # to connect it will wait forever for something to connect. What
                # we do here is check if there is any incoming data on the
                # underlying IP socket used internally by multiprocessing.Listener.
                socket = self.listener._listener._socket
                sel, _, _ = select.select([socket], [], [], 0)
                if len(sel):
                    self.conn = self.listener.accept()
            if self.conn is None:
                return
            conn = self.conn
            global_ns = self.global_ns
            local_ns = self.local_ns
            paused = 1
            while conn and paused != 0:
                if paused >= 0 and not conn.poll():
                    return
                try:
                    job = conn.recv()
                except:
                    self.conn = None
                    break
                jobtype, jobargs = job
                if paused == 1: paused = 0
                try:
                    result = None
                    if jobtype == 'exec':
                        exec_(jobargs, global_ns, local_ns)
                    elif jobtype == 'eval':
                        result = eval(jobargs, global_ns, local_ns)
                    elif jobtype == 'setvar':
                        varname, varval = jobargs
                        local_ns[varname] = varval
                    elif jobtype == 'pause':
                        paused = -1
                    elif jobtype == 'go':
                        paused = 0
                except Exception as e:
                    # if it raised an exception, we return that exception and the
                    # client can then raise it.
                    result = e
                conn.send(result)
        except IOError:
            self.conn = None


class MonitoringClient(object):
    '''
    Used to remotely control (via IP) a running KlustaKwik script

    Initialisation arguments:

    ``server``
        The IP server details, a pair (host, port). If you want to allow control
        only on the one machine (for example, by an IPython shell), leave this
        as ``None`` (which defaults to host='localhost', port=2719). To allow
        remote control, use ('', portnumber).
    ``authkey``
        The authentication key to allow access, change it from 'klustakwik' if you
        are allowing access from outside (otherwise you allow others to run
        arbitrary code on your machine).

    Use a :class:`MonitoringServer` on the run you want to control.

    Has the following methods:

    .. method:: execute(code)

        Executes the specified code in the server process.
        If it raises an
        exception, the server process will catch it and reraise it in the
        client process.

    .. method:: evaluate(code)

        Evaluate the code in the server process and return the result.
        If it raises an
        exception, the server process will catch it and reraise it in the
        client process.

    .. method:: set(name, value)

        Sets the variable ``name`` (a string) to the given value (can be an
        array, etc.). Note that the variable is set in the local namespace, not
        the global one, and so this cannot be used to modify global namespace
        variables. To do that, set a local namespace variable and then
        call :meth:`~MonitoringClient.execute` with an instruction to change
        the global namespace variable.

    .. method:: pause()

        Temporarily stop the script in the server process, continue
        simulation with the :meth:'go' method.

    .. method:: go()

        Continue a script that was paused.

    .. method:: stop()

        Stop a script, equivalent to ``execute('exit(0)')``.

    **Example usage**

    Main script code includes a line like this::

        server = MonitoringServer()

    In an IPython shell you can do something like this::

        client = MonitoringClient()
        print client.evaluate('kk.num_clusters_alive')
   '''
    def __init__(self, server=None, authkey='klustakwik'):
        if multiprocessing is None:
            raise ImportError('Cannot import the required multiprocessing module.')
        if server is None:
            server = ('localhost', 2719)
        self.client = Client(server, authkey=authkey)

    def execute(self, code):
        self.client.send(('exec', code))
        result = self.client.recv()
        if isinstance(result, Exception):
            raise result

    def evaluate(self, code):
        self.client.send(('eval', code))
        result = self.client.recv()
        if isinstance(result, Exception):
            raise result
        return result

    def set(self, name, value):
        self.client.send(('setvar', (name, value)))
        result = self.client.recv()
        if isinstance(result, Exception):
            raise result

    def pause(self):
        self.client.send(('pause', ''))
        self.client.recv()

    def go(self):
        self.client.send(('go', ''))
        self.client.recv()

    def stop(self):
        self.execute('exit(0)')



'''
Some tools for debugging
'''
from numpy import *
import time
import subprocess
import os
from six import iteritems, exec_

__all__ = ['dump_covariance_matrices', 'dump_variable', 'dump_all', 'get_kk_version']


def get_kk_version(version):
    curdir = os.getcwd()
    filedir, _ = os.path.split(__file__)
    os.chdir(filedir)
    try:
        fnull = open(os.devnull, 'w')
        version = 'git-'+subprocess.check_output(['git', 'describe', '--abbrev=8', '--dirty',
                                                       '--always', '--tags'], stderr=fnull).strip().decode('ascii')
    except:
        pass
    os.chdir(curdir)
    return version


def covariance_matrix_dump_callback(kk):
    for cluster, cov in enumerate(kk.covariance):
        msg = 'Covariance matrix for cluster %d:\n' % cluster
        msg += 'Unmasked: %s\nMasked: %s\n' % (cov.unmasked, cov.masked)
        msg += 'Block:\n%s\nDiagonal:\n%s\n' % (cov.block, cov.diagonal)
        kk.log('debug', msg, suffix='covariance_matrices')


def dump_covariance_matrices(kk):
    kk.register_callback(covariance_matrix_dump_callback, 'end_M_step')


class DumpVariableCallback(object):
    def __init__(self, varname, iscode, suffix):
        if not iscode:
            self.varcode = 'kk.%s' % varname
        else:
            self.varcode = varname
        self.suffix = suffix
        self.ns = {}
        exec_('from numpy import *', self.ns)
    def __call__(self, kk):
        self.ns['kk'] = kk
        obj = eval(self.varcode, self.ns)
        msg = str(obj)
        if '\n' in msg:
            msg = '\n'+msg
        kk.log('debug', msg, suffix=self.suffix)


def dump_variable(kk, varname, slot='end_iteration', suffix=None, iscode=False):
    if suffix is None:
        suffix = varname
    kk.register_callback(DumpVariableCallback(varname, iscode, suffix), slot=slot)


class DumpAllCallback(object):
    def __init__(self, slot):
        self.slot = slot
    def __call__(self, kk, *args, **kwds):
        kk.log('debug', 'Dumping variables from slot '+self.slot, suffix=self.slot)
        for i, arg in enumerate(args):
            self.dump_var(kk, 'arg'+str(i), arg)
        for k, v in iteritems(kwds):
            self.dump_var(kk, k, v)
    def dump_var(self, kk, name, val):
        msg = name+' = '
        if isinstance(val, ndarray):
            msg = msg+'\n'+str(val)
        else:
            msg = msg+str(val)
        kk.log('debug', msg, suffix=self.slot)


def dump_all(kk, slot):
    kk.register_callback(DumpAllCallback(slot), slot=slot)



import nose
import os
import klustakwik2

__all__ = ['run']

def run():
    klustakwik2.console_log_level('critical')
    dirname = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    return nose.run(argv=['', dirname,
                          '-c=',  # no config file loading
                          '-I', '^\.',
                          '-I', '^_',
                          '--logging-clear-handlers',
                          '--nologcapture',
                          ])



'''
The numerics module contains all code that is to be compiled. At the moment there are two backends,
the Cython one (cylib) and the Numba one (numbalib). Only the Cython one is complete. Later, when a decision is made
about what to use, we will refactor this code and remove the multiple backends (which isn't maintainable).
'''

from .cylib import *



from .compute_cluster_penalties_cy import do_compute_penalties
from numpy import zeros, amax

__all__ = ['compute_penalties']

def compute_penalties(kk, clusters):
        if clusters is None:
            clusters = kk.clusters
        num_clusters = amax(clusters)+1
        cluster_penalty = zeros(num_clusters)
        ustart = kk.data.unmasked_start
        uend = kk.data.unmasked_end
        penalty_k = kk.penalty_k
        penalty_k_log_n = kk.penalty_k_log_n
        float_num_unmasked = kk.data.float_num_unmasked
        num_spikes = zeros(num_clusters, dtype=int)

        do_compute_penalties(cluster_penalty, num_spikes, clusters,
                             penalty_k, penalty_k_log_n, float_num_unmasked)

        return cluster_penalty
    


from .m_step_cy import *
from six.moves import range

__all__ = ['compute_cluster_mean', 'compute_covariance_matrix']

def get_diagonal(x):
    '''
    Return a writeable view of the diagonal of x
    '''
    return x.reshape(-1)[::x.shape[0]+1]


def compute_cluster_mean(kk, cluster):
    data = kk.data
    num_clusters = len(kk.num_cluster_members)
    num_features = kk.num_features
    
    cluster_mean = numpy.zeros(num_features)
    num_added = numpy.zeros(num_features, dtype=int)
    spikes = kk.get_spikes_in_cluster(cluster)

    prior = 0
    if cluster==kk.mua_cluster:
        prior = kk.mua_point
    elif cluster>=kk.num_special_clusters:
        prior = kk.prior_point

    do_compute_cluster_mean(
                spikes, data.unmasked, data.unmasked_start, data.unmasked_end,
                data.features, data.values_start, data.values_end,
                cluster_mean, num_added, data.noise_mean, prior)

    return cluster_mean


def compute_covariance_matrix(kk, cluster, cluster_mean, cov):
    if cluster<kk.first_gaussian_cluster:
        return
    data = kk.data
    num_cluster_members = kk.num_cluster_members
    num_clusters = len(num_cluster_members)
    num_features = kk.num_features

    block = cov.block
    block_diagonal = get_diagonal(block)

    spikes_in_cluster = kk.spikes_in_cluster
    spikes_in_cluster_offset = kk.spikes_in_cluster_offset
    spike_indices = spikes_in_cluster[spikes_in_cluster_offset[cluster]:spikes_in_cluster_offset[cluster+1]]

    f2m = numpy.zeros(num_features)
    ct = numpy.zeros(num_features)
    
    if kk.use_mua_cluster and cluster==kk.mua_cluster:
        point = kk.mua_point
        do_var_accum_mua(spike_indices, cluster_mean,
                         kk.data.noise_mean, kk.data.noise_variance,
                         cov.unmasked, block,
                         data.unmasked, data.unmasked_start, data.unmasked_end,
                         data.features, data.values_start, data.values_end,
                         f2m, ct, data.correction_terms, num_features,
                         )
    else:
        point = kk.prior_point
        do_var_accum(spike_indices, cluster_mean,
                     kk.data.noise_mean, kk.data.noise_variance,
                     cov.unmasked, block,
                     data.unmasked, data.unmasked_start, data.unmasked_end,
                     data.features, data.values_start, data.values_end,
                     f2m, ct, data.correction_terms, num_features,
                     )
        
    # add correction term for diagonal
    cov.diagonal[:] += len(spike_indices)*data.noise_variance[cov.masked]
    
    # Add prior
    block_diagonal[:] += point*data.noise_variance[cov.unmasked]
    cov.diagonal[:] += point*data.noise_variance[cov.masked]
    
    # Normalise
    factor = 1.0/(num_cluster_members[cluster]+point-1)
    cov.block *= factor
    cov.diagonal *= factor



'''
Cluster data starting from a fet/fmask pair and outputting a clu

Usage:
    kk2_legacy basefname shanknum [param=val, ...]
    
Do not put any spaces in each param=val.
'''

from numpy import *
from klustakwik2 import *

import sys
import time

def main():
    script_params = default_parameters.copy()
    script_params.update(
        drop_last_n_features=0,
        save_clu_every=None,
        run_monitoring_server=False,
        save_all_clu=False,
        debug=True,
        start_from_clu=None,
        use_noise_cluster=True,
        use_mua_cluster=True,
        subset_schedule=None,
        )
    (fname, shank), params = parse_args(2, script_params, __doc__.strip()+'\n',
                                        string_args=set(['start_from_clu']))
    
    drop_last_n_features = params.pop('drop_last_n_features')
    save_clu_every = params.pop('save_clu_every')
    run_monitoring_server = params.pop('run_monitoring_server')
    save_all_clu = params.pop('save_all_clu')
    debug = params.pop('debug')
    num_starting_clusters = params.pop('num_starting_clusters')
    start_from_clu = params.pop('start_from_clu')
    use_noise_cluster = params.pop('use_noise_cluster')
    use_mua_cluster = params.pop('use_mua_cluster')
    subset_schedule = params.pop('subset_schedule')
    
    if subset_schedule is not None:
        if save_clu_every is not None:
            print('Note that intermediate clu files will only be saved for the last subset operation.')
        if not use_noise_cluster:
            print('Must use noise cluster when using subsetting')
            exit(1)

    if debug:
        log_to_file(fname+'.klg.'+shank, 'debug')
    else:
        log_to_file(fname+'.klg.'+shank, 'info')
    log_suppress_hierarchy('klustakwik', inclusive=False)

    start_time = time.time()
    raw_data = load_fet_fmask_to_raw(fname, shank, drop_last_n_features=drop_last_n_features)
    log_message('debug', 'Loading data from .fet and .fmask file took %.2f s' % (time.time()-start_time))
    data = raw_data.to_sparse_data()
    
    log_message('info', 'Number of spikes in data set: '+str(data.num_spikes))
    log_message('info', 'Number of unique masks in data set: '+str(data.num_masks))

    kk = KK(data, use_noise_cluster=use_noise_cluster, use_mua_cluster=use_mua_cluster, **params)
    
    if save_clu_every is not None:
        kk.register_callback(SaveCluEvery(fname, shank, save_clu_every, save_all=save_all_clu))
    if run_monitoring_server:
        kk.register_callback(MonitoringServer())
    
    if start_from_clu is None:
        if subset_schedule is None:
            kk.cluster_mask_starts()
        else:
            kk.cluster_with_subset_schedule(num_starting_clusters, subset_schedule)
    else:
        clusters = load_clu(start_from_clu)
        kk.cluster_from(clusters)
    clusters = kk.clusters
    save_clu(kk, fname, shank)
    
if __name__=='__main__':
    main()
    


#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
#cython: infer_types=True
# distutils: language = c++

import numpy
cimport numpy

from cython cimport integral, floating

from cython.parallel import prange, threadid

from libc.math cimport log

import math
cdef double pi = math.pi

# This function assumes a lower triangular matrix
cpdef trisolve(
            floating[:, :] chol_block,
            floating[:] chol_diagonal,
            integral[:] chol_masked,
            integral[:] chol_unmasked,
            integral num_masked, integral num_unmasked,
            floating[:] x,
            floating[:] root,
            ):
    cdef integral ii, jj, i, j
    cdef floating s
    for ii in range(num_unmasked):
        i = chol_unmasked[ii]
        s = x[i]
        for jj in range(ii):
            j = chol_unmasked[jj]
            s -= chol_block[ii, jj]*root[j]
        root[i] = s/chol_block[ii, ii]
    for ii in range(num_masked):
        i = chol_masked[ii]
        root[i] = x[i]/chol_diagonal[ii]


cpdef do_log_p_assign_computations(
            floating[:] noise_mean,
            floating[:] noise_variance,
            floating[:] cluster_mean,
            floating[:] correction_terms,
            floating[:] log_p_best,
            floating[:] log_p_second_best,
            integral[:] clusters,
            integral[:] clusters_second_best,
            integral[:] old_clusters,
            char full_step,
            floating[:] inv_cov_diag,
            integral[:] unmasked,
            integral[:] ustart,
            integral[:] uend,
            floating[:] features,
            integral[:] vstart,
            integral[:] vend,
            floating[:] root,
            floating[:] f2cm,
            integral num_features, integral num_spikes,
            floating log_addition,
            integral cluster,
            floating[:, :] chol_block,
            floating[:] chol_diagonal,
            integral[:] chol_masked,
            integral[:] chol_unmasked,
            integral n_cpu,
            floating[:] cluster_log_p,
            integral[:] candidates,
            char only_evaluate_current_clusters,
            ):
    cdef integral i, j, ii, jj, p, pp, num_unmasked
    cdef floating mahal, log_p, cur_log_p_best, cur_log_p_second_best, s
    cdef integral chol_num_unmasked = len(chol_unmasked)
    cdef integral chol_num_masked = len(chol_masked)
    cdef integral thread_idx
    cdef integral vo # vector offset for root and f2cm

    for pp in prange(num_spikes, nogil=True, num_threads=n_cpu):
        thread_idx = threadid()
        vo = num_features*thread_idx
        
        if full_step and not only_evaluate_current_clusters:
            p = pp
        else:
            p = candidates[pp]
        
        for i in range(num_features):
            f2cm[i+vo] = noise_mean[i]-cluster_mean[i]
        num_unmasked = uend[p]-ustart[p]
        for ii in range(num_unmasked):
            i = unmasked[ustart[p]+ii]
            j = vstart[p]+ii
            f2cm[i+vo] = features[j]-cluster_mean[i]
        
        # TriSolve step, inlined for Cython OpenMP support
        # This code is adapted from the function above, see there for details
        for ii in range(chol_num_unmasked):
            i = chol_unmasked[ii]
            s = f2cm[i+vo]
            for jj in range(ii):
                j = chol_unmasked[jj]
                s = s-chol_block[ii, jj]*root[j+vo]
            root[i+vo] = s/chol_block[ii, ii]
        for ii in range(chol_num_masked):
            i = chol_masked[ii]
            root[i+vo] = f2cm[i+vo]/chol_diagonal[ii]
        
        mahal = 0
        for i in range(num_features):
            mahal = mahal+root[i+vo]*root[i+vo]
        num_unmasked = uend[p]-ustart[p]
#         for ii in range(num_unmasked):
#             i = unmasked[ustart[p]+ii]
#             mahal = mahal+inv_cov_diag[i]*correction_terms[vstart[p]+ii]
        # the above misses out the contribution of the masked terms. To compute this, we note that
        # correction_terms = noise_variance[i] for masked terms. We can also compute inv_cov_diag but rather
        # than do this, we just mimic the computation above to be more sure of getting it right.
        # However! We don't have access to the set of masked points so we have to invert the list of unmasked points
        jj = -1
        j = -1
        for i in range(num_features):
            if i>j and jj+1<num_unmasked:
                jj = jj+1
                j = unmasked[ustart[p]+jj]
            if i==j:
                mahal = mahal+inv_cov_diag[i]*correction_terms[vstart[p]+jj]
            else:
                mahal = mahal+inv_cov_diag[i]*noise_variance[i]
            
        log_p = mahal/2.0+log_addition
        
        cluster_log_p[p] = log_p

    if only_evaluate_current_clusters:
        return

    for pp in range(num_spikes):

        if full_step:
            p = pp
        else:
            p = candidates[pp]
        
        log_p = cluster_log_p[p]
        
        cur_log_p_best = log_p_best[p]
        cur_log_p_second_best = log_p_second_best[p]
         
        if log_p<cur_log_p_best:
            log_p_second_best[p] = cur_log_p_best
            clusters_second_best[p] = clusters[p]
            log_p_best[p] = log_p
            clusters[p] = cluster
        elif log_p<cur_log_p_second_best:
            log_p_second_best[p] = log_p
            clusters_second_best[p] = cluster



#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
#cython: infer_types=True

import numpy
cimport numpy

from cython cimport integral, floating

from libc.math cimport log

cpdef do_compute_penalties(
            floating[:] cluster_penalty,
            integral[:] num_spikes,
            integral[:] clusters,
            floating penalty_k,
            floating penalty_k_log_n,
            floating[:] float_num_unmasked,
            ):
    cdef integral p, c, total_num_spikes, num_clusters
    cdef floating num_unmasked, num_params, mean_params
    total_num_spikes = len(clusters)
    num_clusters = len(cluster_penalty)
    for p in xrange(total_num_spikes):
        c = clusters[p]
        num_unmasked = float_num_unmasked[p]
        num_params = num_unmasked*(num_unmasked+1)/2.0+num_unmasked+1.0
        cluster_penalty[c] += num_params
        num_spikes[c] += 1
    for c in xrange(num_clusters):
        if num_spikes[c]==0:
            num_spikes[c] = 1
        mean_params = cluster_penalty[c]/num_spikes[c]
        cluster_penalty[c] = (penalty_k*mean_params*2+
                              penalty_k_log_n*mean_params*log(total_num_spikes)/2)
